# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.clock import Clock
from kivy.properties import (NumericProperty, ObjectProperty)
from .util import trigger


class PawnBehavior:
	"""Mix-in class for things in places represented graphically"""
	loc_name = ObjectProperty()
	default_image_paths = ['atlas://rltiles/base.atlas/unseen']
	priority = NumericProperty()
	board = ObjectProperty()

	def __init__(self, **kwargs):
		if 'thing' in kwargs:
			kwargs['proxy'] = kwargs['thing']
			del kwargs['thing']
		if 'proxy' in kwargs:
			kwargs['loc_name'] = kwargs['proxy']['location']
		super().__init__(**kwargs)
		self.register_event_type('on_drop')

	def on_proxy(self, *args):
		self.loc_name = self.proxy['location']

	def on_parent(self, *args):
		if not self.parent:
			Clock.schedule_once(self.on_parent, 0)
			return
		self.board = self.parent.board
		self.bind(loc_name=self._trigger_relocate)
		if self.proxy:
			self._trigger_relocate()

	def finalize(self, initial=True):
		if initial:
			self.loc_name = self.proxy['location']
			self.priority = self.proxy.get('_priority', 0.0)
		self.bind(loc_name=self._trigger_push_location)
		super().finalize(initial)

	def unfinalize(self):
		self.unbind(loc_name=self._trigger_push_location)
		super().unfinalize()

	def pull_from_proxy(self, *args):
		super().pull_from_proxy(*args)
		relocate = False
		if self.loc_name != self.proxy['location']:
			self.loc_name = self.proxy['location']
			relocate = True
		if '_priority' in self.proxy:
			self.priority = self.proxy['_priority']
		if relocate:
			self.relocate()

	def relocate(self, *args):
		if not getattr(
			self, '_finalized', False
		) or not self.parent or not self.proxy or not self.proxy.exists:
			return
		try:
			location = self._get_location_wid()
		except KeyError:
			return
		if location != self.parent:
			if self.parent:
				self.parent.remove_widget(self)
			location.add_widget(self)

	_trigger_relocate = trigger(relocate)

	def on_priority(self, *args):
		if self.proxy['_priority'] != self.priority:
			self.proxy['_priority'] = self.priority
		self.parent.restack()

	def push_location(self, *args):
		if self.proxy['location'] != self.loc_name:
			self.proxy['location'] = self.loc_name

	_trigger_push_location = trigger(push_location)

	def _get_location_wid(self):
		return self.board.spot[self.loc_name]

	def on_touch_up(self, touch):
		if touch.grab_current is not self:
			return False
		for spot in self.board.spot.values():
			if self.collide_widget(spot) and spot.name != self.loc_name:
				new_spot = spot
				break
		else:
			new_spot = None

		self.dispatch('on_drop', new_spot)
		touch.ungrab(self)
		return True

	def on_drop(self, spot):
		parent = self.parent
		if spot:
			self.loc_name = self.proxy['location'] = spot.name
			parent.remove_widget(self)
			spot.add_widget(self)
		else:
			x, y = getattr(self, 'rel_pos', (0, 0))
			self.pos = parent.x + x, parent.y + y



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.properties import NumericProperty, ObjectProperty, OptionProperty
from kivy.lang import Builder
from kivy.uix.boxlayout import BoxLayout
from .util import trigger
from LiSE.character import grid_2d_8graph
from networkx import grid_2d_graph


class GridGeneratorDialog(BoxLayout):
	xval = NumericProperty()
	yval = NumericProperty()
	directions = OptionProperty(None, options=[None, 4, 8])

	def generate(self, engine):
		x = int(self.xval)
		y = int(self.yval)
		if x < 1 or y < 1:
			return False
		elif self.directions == 4:
			# instead, we're running just after game init, before the view is open on it, and we'll make a character ourselves
			engine.add_character('physical', grid_2d_graph(x, y))
			return True
		elif self.directions == 8:
			engine.add_character('physical', grid_2d_8graph(x, y))
			return True
		else:
			return False

	def validate(self):
		return self.directions and int(self.xval) and int(self.yval)


Builder.load_string("""
<GridGeneratorDialog>:
    directions: 4 if but4.state == 'down' else 8
    orientation: 'vertical'
    BoxLayout:
        orientation: 'horizontal'
        MenuIntInput:
            id: input_x
            hint_text: str(root.xval) if root.xval else 'x'
            set_value: root.setter('xval')
        Label:
            text: 'x'
            size_hint_x: 0.1
        MenuIntInput:
            id: input_y
            hint_text: str(root.yval) if root.yval else 'y'
            set_value: root.setter('yval')
    BoxLayout:
        ToggleButton:
            id: but4
            group: 'dir'
            text: '4-way'
            state: 'down'
        ToggleButton:
            id: but8
            group: 'dir'
            text: '8-way'""")



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os
from kivy.app import App
from kivy.properties import BooleanProperty, OptionProperty, ObjectProperty
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.dropdown import DropDown
from kivy.uix.modalview import ModalView
from kivy.uix.textinput import TextInput
from kivy.uix.screenmanager import Screen
from kivy.clock import Clock
from kivy.lang import Builder
from .gen import GridGeneratorDialog


class MenuTextInput(TextInput):
	"""Special text input for setting the branch"""
	set_value = ObjectProperty()

	def __init__(self, **kwargs):
		"""Disable multiline, and bind ``on_text_validate`` to ``on_enter``"""
		kwargs['multiline'] = False
		super().__init__(**kwargs)
		self.bind(on_text_validate=self.on_enter)

	def on_enter(self, *args):
		"""Call the setter and blank myself out so that my hint text shows
        up. It will be the same you just entered if everything's
        working.

        """
		if self.text == '':
			return
		self.set_value(Clock.get_time(), self.text)
		self.text = ''
		self.focus = False

	def on_focus(self, *args):
		"""If I've lost focus, treat it as if the user hit Enter."""
		if not self.focus:
			self.on_enter(*args)

	def on_text_validate(self, *args):
		"""Equivalent to hitting Enter."""
		self.on_enter()


class MenuIntInput(MenuTextInput):
	"""Special text input for setting the turn or tick"""

	def insert_text(self, s, from_undo=False):
		"""Natural numbers only."""
		return super().insert_text(''.join(c for c in s if c in '0123456789'),
									from_undo)


class GeneratorButton(Button):
	pass


class WorldStartConfigurator(BoxLayout):
	"""Give options for how to initialize the world state"""
	grid_config = ObjectProperty()
	generator_type = OptionProperty(None, options=['grid'], allownone=True)
	dismiss = ObjectProperty()
	toggle = ObjectProperty()
	starter = ObjectProperty()
	init_board = ObjectProperty()
	generator_dropdown = ObjectProperty()

	def on_generator_dropdown(self, *args):

		def select_txt(btn):
			self.generator_dropdown.select(btn.text)

		for opt in ['None', 'Grid']:
			self.generator_dropdown.add_widget(
				GeneratorButton(text=opt, on_release=select_txt))
		self.generator_dropdown.bind(on_select=self.select_generator_type)

	def select_generator_type(self, instance, value):
		self.ids.drop.text = value
		if value == 'None':
			self.ids.controls.clear_widgets()
			self.generator_type = None
		elif value == 'Grid':
			self.ids.controls.clear_widgets()
			self.ids.controls.add_widget(self.grid_config)
			self.grid_config.size = self.ids.controls.size
			self.grid_config.pos = self.ids.controls.pos
			self.generator_type = 'grid'

	def start(self, *args):
		if self.generator_type == 'grid':
			if self.grid_config.validate():
				engine = self.starter()
				self.grid_config.generate(engine)
				self.init_board()
				self.toggle()
				self.dismiss()
			else:
				# TODO show error
				return
		else:
			self.starter()
			self.init_board()
			self.toggle()
			self.dismiss()


class DirPicker(Screen):
	toggle = ObjectProperty()

	def open(self, path, *args):
		App.get_running_app().starting_dir = os.path.abspath('.')
		os.chdir(path)
		if 'world.db' not in os.listdir(path):
			# TODO show a configurator, accept cancellation, extract init params
			if not hasattr(self, 'config_popover'):
				self.config_popover = ModalView()
				self.configurator = WorldStartConfigurator(
					grid_config=GridGeneratorDialog(),
					dismiss=self.config_popover.dismiss,
					toggle=self.toggle,
					generator_dropdown=DropDown())
				self.config_popover.add_widget(self.configurator)
			self.config_popover.open()
			return
		App.get_running_app().start_subprocess()
		App.get_running_app().init_board()
		self.toggle()


Builder.load_string("""
#: import os os
<GeneratorButton>:
    size_hint_y: None
    height: self.texture_size[1] + 10
<WorldStartConfigurator>:
    orientation: 'vertical'
    init_board: app.init_board
    starter: app.start_subprocess
    Label:
        text: 'Generate an initial map?'
    Button:
        id: drop
        text: 'None'
        on_release: root.generator_dropdown.open(drop)
    Widget:
        id: controls
        size_hint_y: None
        height: 200
    BoxLayout:
        orientation: 'horizontal'
        Button:
            text: 'OK'
            on_release:
                root.start()
        Button:
            text: 'Cancel'
            on_release:
                controls.clear_widgets()
                controls.size_hint_y = 0
                root._trigger_layout()
                root.dismiss()
<DirPicker>:
    name: 'mainmenu'
    start: app.start_subprocess
    init_board: app.init_board
    BoxLayout:
        orientation: 'vertical'
        Label:
            text: 'Pick a directory to create or load a simulation in'
            size_hint_y: None
        FileChooserListView:
            id: filechooser
            path: os.getcwd()
        Button:
            text: 'Work here'
            size_hint_y: 0.1
            on_release: root.open(filechooser.path)
""")



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from .pallet import Pallet, PalletBox
from .kivygarden.texturestack import ImageStack
from kivy.clock import Clock
from kivy.lang import Builder
from kivy.logger import Logger
from kivy.properties import (ListProperty, NumericProperty, ObjectProperty,
								StringProperty)
from kivy.uix.label import Label
from kivy.uix.scrollview import ScrollView
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.screenmanager import Screen
from .util import trigger

# TODO: let the user import their own sprite art


class SpriteSelector(BoxLayout):
	prefix = StringProperty()
	pallets = ListProperty()
	imgpaths = ListProperty([])
	default_imgpaths = ListProperty()
	preview = ObjectProperty()

	def on_prefix(self, *args):
		if 'textbox' not in self.ids:
			Clock.schedule_once(self.on_prefix, 0)
			return
		self.ids.textbox.text = self.prefix

	def on_imgpaths(self, *args):
		if not self.preview:
			Logger.debug("SpriteSelector: no preview")
			Clock.schedule_once(self.on_imgpaths, 0)
			return
		if hasattr(self, '_imgstack'):
			self.preview.remove_widget(self._imgstack)
		self._imgstack = ImageStack(paths=self.imgpaths,
									x=self.preview.center_x - 16,
									y=self.preview.center_y - 16)
		self.preview.add_widget(self._imgstack)

	def on_pallets(self, *args):
		for pallet in self.pallets:
			pallet.fbind('selection', self._upd_imgpaths)

	def _upd_imgpaths(self, *args):
		imgpaths = []
		for pallet in self.pallets:
			if pallet.selection:
				for selected in pallet.selection:
					imgpaths.append('atlas://{}/{}'.format(
						pallet.filename, selected.text))
		self.imgpaths = imgpaths if imgpaths else self.default_imgpaths


class SpriteBuilder(ScrollView):
	prefix = StringProperty()
	imgpaths = ListProperty()
	default_imgpaths = ListProperty()
	data = ListProperty()
	labels = ListProperty()
	pallets = ListProperty()

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(data=self._trigger_update)

	def update(self, *args):
		if self.data is None:
			return
		if not self.canvas:
			Clock.schedule_once(self.update, 0)
			return
		if not hasattr(self, '_palbox'):
			self._palbox = PalletBox(orientation='vertical', size_hint_y=None)
			self.add_widget(self._palbox)
		else:
			self._palbox.clear_widgets()
		if hasattr(self._palbox, '_bound_width'):
			for uid in self._palbox._bound_width:
				self._palbox.unbind_uid('width', uid)
			del self._palbox._bound_width
		self.labels = []
		for pallet in self.pallets:
			if hasattr(pallet, '_bound_minimum_height'):
				pallet.unbind_uid('minimum_height',
									pallet._bound_minimum_height)
				del pallet._bound_minimum_height
			if hasattr(pallet, '_bound_height'):
				pallet.unbind_uid('height', pallet._bound_height)
				del pallet._bound_height
		self.pallets = []
		for (text, filename) in self.data:
			label = Label(text=text, size_hint=(None, None), halign='center')
			label.texture_update()
			label.height = label.texture.height
			label.width = self._palbox.width
			pallet = Pallet(filename=filename, size_hint=(None, None))
			pallet.width = self._palbox.width
			self._palbox._bound_width = [
				self._palbox.fbind('width', label.setter('width')),
				self._palbox.fbind('width', pallet.setter('width'))
			]
			pallet.height = pallet.minimum_height
			pallet._bound_minimum_height = pallet.fbind(
				'minimum_height', pallet.setter('height')),
			pallet._bound_height = pallet.fbind('height',
												self._trigger_reheight)
			self.labels.append(label)
			self.pallets.append(pallet)
		n = len(self.labels)
		assert (n == len(self.pallets))
		for i in range(0, n):
			self._palbox.add_widget(self.labels[i])
			self._palbox.add_widget(self.pallets[i])

	_trigger_update = trigger(update)

	def reheight(self, *args):
		self._palbox.height = sum(wid.height
									for wid in self.labels + self.pallets)

	_trigger_reheight = trigger(reheight)


class SpriteDialog(BoxLayout):
	toggle = ObjectProperty()
	prefix = StringProperty()
	imgpaths = ListProperty()
	default_imgpaths = ListProperty()
	data = ListProperty()
	pallet_box_height = NumericProperty()

	def pressed(self):
		self.prefix = self.ids.selector.prefix
		self.imgpaths = self.ids.selector.imgpaths
		self.toggle()


class PawnConfigDialog(SpriteDialog):
	pass


class SpotConfigDialog(SpriteDialog):
	pass


class PawnConfigScreen(Screen):
	toggle = ObjectProperty()
	data = ListProperty()
	imgpaths = ListProperty()


class SpotConfigScreen(Screen):
	toggle = ObjectProperty()
	data = ListProperty()
	imgpaths = ListProperty()


Builder.load_string("""
<SpriteDialog>:
    orientation: 'vertical'
    SpriteBuilder:
        id: builder
        prefix: root.prefix
        default_imgpaths: root.default_imgpaths
        imgpaths: root.imgpaths
        data: root.data
    SpriteSelector:
        id: selector
        textbox: textbox
        size_hint_y: 0.1
        prefix: root.prefix
        default_imgpaths: root.default_imgpaths
        imgpaths: root.imgpaths
        pallets: builder.pallets
        preview: preview
        TextInput:
            id: textbox
            multiline: False
            write_tab: False
            hint_text: 'Enter name prefix'
        Widget:
            id: preview
        Button:
            text: 'OK'
            on_release: root.pressed()
<PawnConfigScreen>:
    name: 'pawncfg'
    imgpaths: dialog.imgpaths
    PawnConfigDialog:
        id: dialog
        toggle: root.toggle
        default_imgpaths: ['atlas://rltiles/base/unseen']
        data: root.data
<SpotConfigScreen>:
    name: 'spotcfg'
    imgpaths: dialog.imgpaths
    SpotConfigDialog:
        id: dialog
        toggle: root.toggle
        default_imgpaths: ['atlas://rltiles/floor/floor-stone']
        data: root.data
""")



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Generic dialog boxes and menus, for in front of a Board

"""
from functools import partial
from kivy.properties import (DictProperty, ListProperty, ObjectProperty,
								StringProperty, NumericProperty,
								VariableListProperty)
from kivy.core.text import DEFAULT_FONT
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.widget import Widget
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.scrollview import ScrollView
from kivy.clock import Clock
from kivy.lang import Builder
from kivy.logger import Logger


class Box(Widget):
	padding = VariableListProperty(6)
	border = VariableListProperty(4)
	font_size = StringProperty('15sp')
	font_name = StringProperty(DEFAULT_FONT)
	background = StringProperty()
	background_color = VariableListProperty([1, 1, 1, 1])
	foreground_color = VariableListProperty([0, 0, 0, 1])


class ScrollableLabel(ScrollView):
	font_size = StringProperty('15sp')
	font_name = StringProperty(DEFAULT_FONT)
	color = VariableListProperty([0, 0, 0, 1])
	line_spacing = NumericProperty(0)
	text = StringProperty()


class MessageBox(Box):
	"""Looks like a TextInput but doesn't accept any input.

    Does support styled text with BBcode.

    """
	line_spacing = NumericProperty(0)
	text = StringProperty()


class DialogMenu(Box):
	"""Some buttons that make the game do things.

    Set ``options`` to a list of pairs of ``(text, function)`` and the
    menu will be populated with buttons that say ``text`` that call
    ``function`` when pressed.

    """
	options = ListProperty()
	"""List of pairs of (button_text, callable)"""

	def _set_sv_size(self, *args):
		self._sv.width = self.width - self.padding[0] - self.padding[2]
		self._sv.height = self.height - self.padding[1] - self.padding[3]

	def _set_sv_pos(self, *args):
		self._sv.x = self.x + self.padding[0]
		self._sv.y = self.y + self.padding[3]

	def on_options(self, *args):
		if not hasattr(self, '_sv'):
			self._sv = ScrollView(size=self.size, pos=self.pos)
			self.bind(size=self._set_sv_size, pos=self._set_sv_pos)
			layout = BoxLayout(orientation='vertical')
			self._sv.add_widget(layout)
		else:
			layout = self._sv.children[0]
			layout.clear_widgets()
		for txt, part in self.options:
			if not callable(part):
				raise TypeError("Menu options must be callable")
			layout.add_widget(
				Button(text=txt,
						on_release=part,
						font_name=self.font_name,
						font_size=self.font_size))


class Dialog(BoxLayout):
	"""MessageBox with a DialogMenu beneath it.

    Set the properties ``message_kwargs`` and ``menu_kwargs``,
    respectively, to control them -- but you probably want
    to do that by returning a pair of dicts from an action
    in LiSE.

    """
	message_kwargs = DictProperty({})
	menu_kwargs = DictProperty({})

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(message_kwargs=self._propagate_msg_kwargs,
					menu_kwargs=self._propagate_menu_kwargs)
		self._propagate_msg_kwargs()
		self._propagate_menu_kwargs()

	def _propagate_msg_kwargs(self, *args):
		if 'msg' not in self.ids:
			Clock.schedule_once(self._propagate_msg_kwargs, 0)
			return
		kw = dict(self.message_kwargs)
		kw.setdefault('background',
						'atlas://data/images/defaulttheme/textinput')
		for k, v in kw.items():
			setattr(self.ids.msg, k, v)

	def _propagate_menu_kwargs(self, *args):
		if 'menu' not in self.ids:
			Clock.schedule_once(self._propagate_menu_kwargs, 0)
			return
		kw = dict(self.menu_kwargs)
		kw.setdefault(
			'background',
			'atlas://data/images/defaulttheme/vkeyboard_background')
		for k, v in kw.items():
			setattr(self.ids.menu, k, v)


class DialogLayout(FloatLayout):
	"""A layout, normally empty, that can generate dialogs

    To make dialogs, set my ``todo`` property to a list. It may contain:

    * Strings, which will be displayed with an "OK" button to dismiss them
    * Lists of pairs of strings and callables, which generate buttons with the string
    on them that, when clicked, call the callable
    * Lists of pairs of dictionaries, which are interpreted as keyword arguments
    to :class:`MessageBox` and :class:`DialogMenu`

    In place of a callable you can use the name of a function in my ``usermod``,
    a Python module given by name. I'll import it when I need it.

    Needs to be instantiated with a LiSE ``engine`` -- probably an
    :class:`EngineProxy`.

    """
	dialog = ObjectProperty()
	engine = ObjectProperty()
	todo = ListProperty()
	idx = NumericProperty()
	usermod = StringProperty('user')
	userpkg = StringProperty(None, allownone=True)

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.dialog = Dialog()

	def on_engine(self, *args):
		todo = self.engine.universal.get('last_result')
		if isinstance(todo, list):
			self.todo = todo
		else:
			self.todo = []
		self.idx = self.engine.universal.get('last_result_idx', 0)
		self.engine.universal.connect(self._pull)
		if self.todo:
			self.advance_dialog()

	def _pull(self, *args, key, value):
		if key == 'last_result':
			self.todo = value if value and isinstance(value, list) else []
		elif key == 'last_result_idx':
			self.idx = value if value and isinstance(value, int) else 0

	def on_idx(self, *args):
		lidx = self.engine.universal.get('last_result_idx')
		if lidx is not None and lidx != self.idx:
			self.engine.universal['last_result_idx'] = self.idx

	def advance_dialog(self, *args):
		"""Try to display the next dialog described in my ``todo``."""
		self.clear_widgets()
		try:
			self._update_dialog(self.todo[self.idx])
		except IndexError:
			pass

	def _update_dialog(self, diargs, **kwargs):
		if diargs is None:
			Logger.debug("DialogLayout: null dialog")
			return
		dia = self.dialog
		# Simple text dialogs just tell the player something and let them click OK
		if isinstance(diargs, str):
			dia.message_kwargs = {'text': diargs}
			dia.menu_kwargs = {'options': [('OK', self._trigger_ok)]}
		# List dialogs are for when you need the player to make a choice and don't care much
		# about presentation
		elif isinstance(diargs, list):
			dia.message_kwargs = {'text': 'Select from the following:'}
			dia.menu_kwargs = {
				'options': list(map(self._munge_menu_option, diargs))
			}
		# For real control of the dialog, you need a pair of dicts --
		# the 0th describes the message shown to the player, the 1th
		# describes the menu below
		elif isinstance(diargs, tuple):
			if len(diargs) != 2:
				# TODO more informative error
				raise TypeError('Need a tuple of length 2')
			msgkwargs, mnukwargs = diargs
			if isinstance(msgkwargs, dict):
				dia.message_kwargs = msgkwargs
			elif isinstance(msgkwargs, str):
				dia.message_kwargs['text'] = msgkwargs
			else:
				raise TypeError("Message must be dict or str")
			if isinstance(mnukwargs, dict):
				mnukwargs['options'] = list(
					map(self._munge_menu_option, mnukwargs['options']))
				dia.menu_kwargs = mnukwargs
			elif isinstance(mnukwargs, list) or isinstance(mnukwargs, tuple):
				dia.menu_kwargs['options'] = list(
					map(self._munge_menu_option, mnukwargs))
			else:
				raise TypeError("Menu must be dict or list")
		else:
			raise TypeError("Don't know how to turn {} into a dialog".format(
				type(diargs)))
		self.add_widget(dia)

	def ok(self, *args, cb=None):
		"""Clear dialog widgets, call ``cb`` if provided, and advance the dialog queue"""
		self.clear_widgets()
		if cb:
			cb()
		self.idx += 1
		self.advance_dialog()

	def _trigger_ok(self, *args, cb=None):
		part = partial(self.ok, cb=cb)
		if hasattr(self, '_scheduled_ok'):
			Clock.unschedule(self._scheduled_ok)
		self._scheduled_ok = Clock.schedule_once(part)

	def _lookup_func(self, funcname):
		from importlib import import_module
		if not hasattr(self, '_usermod'):
			self._usermod = import_module(self.usermod, self.userpkg)
		return getattr(self.usermod, funcname)

	def _munge_menu_option(self, option):
		if not isinstance(option, tuple):
			raise TypeError
		name, func = option
		if func is None:
			return name, self._trigger_ok
		if callable(func):
			return name, partial(self._trigger_ok, cb=func)
		if isinstance(func, tuple):
			fun = func[0]
			if isinstance(fun, str):
				fun = self._lookup_func(fun)
			args = func[1]
			if len(func) == 3:
				kwargs = func[2]
				func = partial(fun, *args, **kwargs)
			else:
				func = partial(fun, *args)
		if isinstance(func, str):
			func = self._lookup_func(func)
		return name, partial(self._trigger_ok, cb=func)


Builder.load_string("""
<Box>:
    canvas.before:
        Color:
            rgba: self.background_color
        BorderImage:
            border: self.border
            pos: self.pos
            size: self.size
            source: self.background
        Color:
            rgba: 1, 1, 1, 1
<ScrollableLabel>:
    Label:
        size_hint_y: None
        height: self.texture_size[1]
        text_size: self.width, None
        text: root.text
        color: root.color
<MessageBox>:
    ScrollableLabel:
        x: root.x + root.padding[0]
        y: root.y + root.padding[3]
        width: root.width - root.padding[2]
        height: root.height - root.padding[1]
        text: root.text
        color: root.foreground_color
<Dialog>:
    orientation: 'vertical'
    pos_hint: {'x': 0, 'y': 0}
    size_hint: 1, 0.3
    MessageBox:
        id: msg
    DialogMenu:
        id: menu
""")
